
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.naming.Context;
import javax.naming.InitialContext;

/**
* This class checks for message loss, duplicates and out of order messages.
* This consumer (and producer) should be connected to a cluster with rapid membership
* changes to test for the above properties.
*/
public class FailoverConsumer
{
   private int numMessages = 5000;    
   private static final String CLASS = "FailoverConsumer";


   public FailoverConsumer()
   {
       numMessages = Integer.getInteger("msg_count", 5000);
   }
   
   /**
    * Run the message consumer example.
    * @param args Command line arguments.
    */
   public static void main(String[] args)
   {
       FailoverConsumer syncConsumer = new FailoverConsumer();
       syncConsumer.runTest();
   }

   /**
    * Start the example.
    */
   private void runTest()
   {
       try
       {
           Properties properties = new Properties();
           properties.load(this.getClass().getResourceAsStream("failover.properties"));

           Context ctx = new InitialContext(properties);

           Destination destination = Boolean.getBoolean("persistent")? 
                   (Destination)ctx.lookup("directDuraQueue"):(Destination)ctx.lookup("directQueue");

           ConnectionFactory conFac = (ConnectionFactory)ctx.lookup("qpidConnectionfactory");
           Connection connection = conFac.createConnection();
         
           System.out.println(
                   CLASS + ": Setting an ExceptionListener on the connection as sample uses a MessageConsumer");

           connection.setExceptionListener(new ExceptionListener()
           {
               public void onException(JMSException jmse)
               {
                   System.err.println(CLASS + ": The sample received an exception through the ExceptionListener");
                   System.exit(0);
               }
           });

           System.out.println(CLASS + ": Creating a non-transacted, auto-acknowledged session");
           Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);

           System.out.println(CLASS + ": Creating a MessageConsumer");
           MessageConsumer messageConsumer = session.createConsumer(destination);
           
           System.out.println(CLASS + ": Starting connection so MessageConsumer can receive messages");
           connection.start();

           Message message = null;
           boolean end = false;
          
           int rcvdSeq = 0;
           int msgCount = 0;
           boolean duplicateMsgs = false;
           boolean outOfOrder = false;
           List<Integer> duplicateMessages = new ArrayList<Integer>();
           List<Integer> outOfOrderMessages = new ArrayList<Integer>();
           while (!end)
           {
           
               message = messageConsumer.receive();
               String text = ((TextMessage) message).getText();
               
               if (text.equals("That's all, folks!"))
               {
                   System.out.println(CLASS + ": Received final message " + text);
                   end = true;
               }
               else
               {
                   int seq = message.getIntProperty("seq");                   
                   if (seq <= rcvdSeq)
                   {
                       System.out.println(CLASS + ": Received a ***duplicate*** message:  " + text);
                       duplicateMessages.add(seq);
                       duplicateMsgs = true;
                   }
                   else if (seq == rcvdSeq + 1)
                   {
                       rcvdSeq++;
                       msgCount++;
                       System.out.println(CLASS + ": Received  message:  " + text);
                   }
                   else
                   {
                       outOfOrder = true;
                       msgCount++;
                       outOfOrderMessages.add(seq);
                       System.out.println(CLASS + ": Received an ***out of order*** message:  " + text);
                   }
               }
           
           }          
           
           if (duplicateMsgs)
           {
               System.out.println(CLASS + ": The following messages were duplicated");
               for (int x:duplicateMessages)
               {
                   System.out.println("message" + x);
               }
           }
           
           if (outOfOrder)
           {
               System.out.println(CLASS + ": There were " + outOfOrderMessages.size() + " messages were out of order");
               /*for (int x:duplicateMessages)
               {
                   System.out.println("message" + x);
               }*/
               
           }
           
           if (msgCount != numMessages)
           {
               System.out.println(CLASS + 
                       ": There were lost messages. Expected " + 
                       numMessages + " but received only " + msgCount);
           }
               
           System.out.println(CLASS + ": Closing connection");
           connection.close();
           System.out.println(CLASS + ": Closing JNDI context");
           ctx.close();
           
       }
       catch (Exception exp)
       {           
           System.err.println(CLASS + ": Caught an Exception: " + exp);
           exp.printStackTrace();
       }
   }
}
